from __future__ import division
from itertools import product
import collections

import numpy as np
from scipy.stats import multivariate_normal as mvn
from scipy.stats import norm
import scipy.spatial as spatial
import matplotlib.pyplot as plt

from sklearn.base import TransformerMixin

__all__ = ["PersImage"]

class PersImage(TransformerMixin):
    """ Initialize a persistence image generator.

    Parameters
    -----------

    pixels : pair of ints like (int, int)
        Tuple representing number of pixels in return image along x and y axis.
    spread : float
        Standard deviation of gaussian kernel
    specs : dict
        Parameters for shape of image with respect to diagram domain. This is used if you would like images to have a particular range. Shaped like 
        ::
        
            {
                "maxBD": float,
                "minBD": float
            }

    kernel_type : string or ...
        TODO: Implement this feature.
        Determine which type of kernel used in the convolution, or pass in custom kernel. Currently only implements Gaussian.
    weighting_type : string or ...
        TODO: Implement this feature.
        Determine which type of weighting function used, or pass in custom weighting function.
        Currently only implements linear weighting.


    Usage
    ------


    """

    def __init__(
        self,
        pixels=(20, 20),
        spread=None,
        specs=None,
        kernel_type="gaussian",
        weighting_type="linear",
        verbose=True,
    ):

        self.specs = specs
        self.kernel_type = kernel_type
        self.weighting_type = weighting_type
        self.spread = spread
        self.nx, self.ny = pixels

        if verbose:
            print(
                'PersImage(pixels={}, spread={}, specs={}, kernel_type="{}", weighting_type="{}")'.format(
                    pixels, spread, specs, kernel_type, weighting_type
                )
            )

    def transform(self, diagrams):
        """ Convert diagram or list of diagrams to a persistence image.

        Parameters
        -----------

        diagrams : list of or singleton diagram, list of pairs. [(birth, death)]
            Persistence diagrams to be converted to persistence images. It is assumed they are in (birth, death) format. Can input a list of diagrams or a single diagram.

        """
        # if diagram is empty, return empty image
        if len(diagrams) == 0:
            return np.zeros((self.nx, self.ny))
        # if first entry of first entry is not iterable, then diagrams is singular and we need to make it a list of diagrams
        try:
            singular = not isinstance(diagrams[0][0], collections.Iterable)
        except IndexError:
            singular = False

        if singular:
            diagrams = [diagrams]

        dgs = [np.copy(diagram, np.float64) for diagram in diagrams]
        landscapes = [PersImage.to_landscape(dg) for dg in dgs]

        if not self.specs:
            self.specs = {
                "maxBD": np.max([np.max(np.vstack((landscape, np.zeros((1, 2))))) 
                                 for landscape in landscapes] + [0]),
                "minBD": np.min([np.min(np.vstack((landscape, np.zeros((1, 2))))) 
                                 for landscape in landscapes] + [0]),
            }
        imgs = [self._transform(dgm) for dgm in landscapes]

        # Make sure we return one item.
        if singular:
            imgs = imgs[0]

        return imgs

    def _transform(self, landscape):
        # Define an NxN grid over our landscape
        maxBD = self.specs["maxBD"]
        minBD = min(self.specs["minBD"], 0)  # at least show 0, maybe lower

        # Same bins in x and y axis
        dx = maxBD / (self.ny)
        xs_lower = np.linspace(minBD, maxBD, self.nx)
        xs_upper = np.linspace(minBD, maxBD, self.nx) + dx

        ys_lower = np.linspace(0, maxBD, self.ny)
        ys_upper = np.linspace(0, maxBD, self.ny) + dx

        weighting = self.weighting(landscape)

        # Define zeros
        img = np.zeros((self.nx, self.ny))

        # Implement this as a `summed-area table` - it'll be way faster
        spread = self.spread if self.spread else dx
        for point in landscape:
            x_smooth = norm.cdf(xs_upper, point[0], spread) - norm.cdf(
                xs_lower, point[0], spread
            )
            y_smooth = norm.cdf(ys_upper, point[1], spread) - norm.cdf(
                ys_lower, point[1], spread
            )
            img += np.outer(x_smooth, y_smooth) * weighting(point)
        img = img.T[::-1]
        return img

    def weighting(self, landscape=None):
        """ Define a weighting function, 
                for stability results to hold, the function must be 0 at y=0.    
        """

        # TODO: Implement a logistic function
        # TODO: use self.weighting_type to choose function

        if landscape is not None:
            if len(landscape) > 0:
                maxy = np.max(landscape[:, 1])
            else: 
                maxy = 1

        def linear(interval):
            # linear function of y such that f(0) = 0 and f(max(y)) = 1
            d = interval[1]
            return (1 / maxy) * d if landscape is not None else d

        def pw_linear(interval):
            """ This is the function defined as w_b(t) in the original PI paper

                Take b to be maxy/self.ny to effectively zero out the bottom pixel row
            """

            t = interval[1]
            b = maxy / self.ny

            if t <= 0:
                return 0
            if 0 < t < b:
                return t / b
            if b <= t:
                return 1

        return linear

    def kernel(self, spread=1):
        """ This will return whatever kind of kernel we want to use.
            Must have signature (ndarray size NxM, ndarray size 1xM) -> ndarray size Nx1
        """
        # TODO: use self.kernel_type to choose function

        def gaussian(data, pixel):
            return mvn.pdf(data, mean=pixel, cov=spread)

        return gaussian

    @staticmethod
    def to_landscape(diagram):
        """ Convert a diagram to a landscape
            (b,d) -> (b, d-b)
        """
        diagram[:, 1] -= diagram[:, 0]

        return diagram

    def show(self, imgs, ax=None):
        """ Visualize the persistence image

        """

        ax = ax or plt.gca()

        if type(imgs) is not list:
            imgs = [imgs]

        for i, img in enumerate(imgs):
            ax.imshow(img, cmap=plt.get_cmap("plasma"))
            ax.axis("off")



"""

    Implementation of the bottleneck distance

    Author: Chris Tralie

"""

import numpy as np

from bisect import bisect_left
from hopcroftkarp import HopcroftKarp

__all__ = ["bottleneck"]


def bottleneck(dgm1, dgm2, matching=False):
    """
    Perform the Bottleneck distance matching between persistence diagrams.
    Assumes first two columns of S and T are the coordinates of the persistence
    points, but allows for other coordinate columns (which are ignored in
    diagonal matching).

    See the `distances` notebook for an example of how to use this.

    Parameters
    -----------
    dgm1: Mx(>=2) 
        array of birth/death pairs for PD 1
    dgm2: Nx(>=2) 
        array of birth/death paris for PD 2
    matching: bool, default False
        if True, return matching infromation and cross-similarity matrix

    Returns
    --------

    d: float
        bottleneck distance between dgm1 and dgm2
    (matching, D): Only returns if `matching=True`
        (tuples of matched indices, (N+M)x(N+M) cross-similarity matrix)
    """

    return_matching = matching

    S = np.array(dgm1)
    S = S[np.isfinite(S[:, 1]), :]
    T = np.array(dgm2)
    T = T[np.isfinite(T[:, 1]), :]

    N = S.shape[0]
    M = T.shape[0]

    # Step 1: Compute CSM between S and T, including points on diagonal
    # L Infinity distance
    Sb, Sd = S[:, 0], S[:, 1]
    Tb, Td = T[:, 0], T[:, 1]
    D1 = np.abs(Sb[:, None] - Tb[None, :])
    D2 = np.abs(Sd[:, None] - Td[None, :])
    DUL = np.maximum(D1, D2)

    # Put diagonal elements into the matrix, being mindful that Linfinity
    # balls meet the diagonal line at a diamond vertex
    D = np.zeros((N + M, N + M))
    D[0:N, 0:M] = DUL
    UR = np.max(D) * np.ones((N, N))
    np.fill_diagonal(UR, 0.5 * (S[:, 1] - S[:, 0]))
    D[0:N, M::] = UR
    UL = np.max(D) * np.ones((M, M))
    np.fill_diagonal(UL, 0.5 * (T[:, 1] - T[:, 0]))
    D[N::, 0:M] = UL

    # Step 2: Perform a binary search + Hopcroft Karp to find the
    # bottleneck distance
    N = D.shape[0]
    ds = np.sort(np.unique(D.flatten()))
    bdist = ds[-1]
    matching = {}
    while len(ds) >= 1:
        idx = 0
        if len(ds) > 1:
            idx = bisect_left(range(ds.size), int(ds.size / 2))
        d = ds[idx]
        graph = {}
        for i in range(N):
            graph["%s" % i] = {j for j in range(N) if D[i, j] <= d}
        res = HopcroftKarp(graph).maximum_matching()
        if len(res) == 2 * N and d < bdist:
            bdist = d
            matching = res
            ds = ds[0:idx]
        else:
            ds = ds[idx + 1::]

    if return_matching:
        matchidx = [(i, matching["%i" % i]) for i in range(N)]
        return bdist, (matchidx, D)
    else:
        return bdist



import numpy as np
from sklearn import metrics
from scipy import optimize

__all__ = ["wasserstein"]


def wasserstein(dgm1, dgm2, matching=False):
    """
    Perform the Wasserstein distance matching between persistence diagrams.
    Assumes first two columns of dgm1 and dgm2 are the coordinates of the persistence
    points, but allows for other coordinate columns (which are ignored in
    diagonal matching).

    See the `distances` notebook for an example of how to use this.

    Parameters
    ------------

    dgm1: Mx(>=2) 
        array of birth/death pairs for PD 1
    dgm2: Nx(>=2) 
        array of birth/death paris for PD 2
    matching: bool, default False
        if True, return matching infromation and cross-similarity matrix

    Returns 
    ---------

    d: float
        Wasserstein distance between dgm1 and dgm2
    (matching, D): Only returns if `matching=True`
        (tuples of matched indices, (N+M)x(N+M) cross-similarity matrix)

    """

    # Step 1: Compute CSM between S and dgm2, including points on diagonal
    N = dgm1.shape[0]
    M = dgm2.shape[0]
    # Handle the cases where there are no points in the diagrams
    if N == 0:
        dgm1 = np.array([[0, 0]])
        N = 1
    if M == 0:
        dgm2 = np.array([[0, 0]])
        M = 1
    DUL = metrics.pairwise.pairwise_distances(dgm1, dgm2)

    # Put diagonal elements into the matrix
    # Rotate the diagrams to make it easy to find the straight line
    # distance to the diagonal
    cp = np.cos(np.pi/4)
    sp = np.sin(np.pi/4)
    R = np.array([[cp, -sp], [sp, cp]])
    dgm1 = dgm1[:, 0:2].dot(R)
    dgm2 = dgm2[:, 0:2].dot(R)
    D = np.zeros((N+M, N+M))
    D[0:N, 0:M] = DUL
    UR = np.max(D)*np.ones((N, N))
    np.fill_diagonal(UR, dgm1[:, 1])
    D[0:N, M:M+N] = UR
    UL = np.max(D)*np.ones((M, M))
    np.fill_diagonal(UL, dgm2[:, 1])
    D[N:M+N, 0:M] = UL

    # Step 2: Run the hungarian algorithm
    matchi, matchj = optimize.linear_sum_assignment(D)
    matchdist = np.sum(D[matchi, matchj])

    if matching:
        matchidx = [(i, j) for i, j in zip(matchi, matchj)]
        return matchdist, (matchidx, D)

    return matchdist



import pytest

import numpy as np
from persim import PersImage


def test_landscape():
    bds = np.array([[1, 1], [1, 2]])

    ldsp = PersImage.to_landscape(bds)

    np.testing.assert_array_equal(ldsp, [[1, 0], [1, 1]])


def test_integer_diagrams():
    """ This test is inspired by gh issue #3 by gh user muszyna25.

    Integer diagrams return nan values.

    This does not work: dgm = [[0, 2], [0, 6], [0, 8]];

    This one works fine: dgm = [[0.0, 2.0], [0.0, 6.0], [0.0, 8.0]];

    """

    dgm = [[0, 2], [0, 6], [0, 8]]
    dgm2 = [[0.0, 2.0], [0.0, 6.0], [0.0, 8.0]]
    pim = PersImage()
    res = pim.transform(dgm2)
    res2 = pim.transform(dgm)
    np.testing.assert_array_equal(res, res2)

class TestEmpty:
    def test_empty_diagram(self):
        dgm = np.zeros((0, 2))
        pim = PersImage(pixels = (10, 10))
        res = pim.transform(dgm)
        assert np.all(res == np.zeros((10, 10)))

    def test_empyt_diagram_list(self):
        dgm1 = [np.array([[2, 3]]), 
                np.zeros((0, 2))]
        pim1 = PersImage(pixels = (10, 10))
        res1 = pim1.transform(dgm1)
        assert np.all(res1[1] == np.zeros((10, 10)))

        dgm2 = [np.zeros((0, 2)), 
                np.array([[2, 3]])]
        pim2 = PersImage(pixels = (10, 10))
        res2 = pim2.transform(dgm2)
        assert np.all(res2[0] == np.zeros((10, 10)))

        dgm3 = [np.zeros((0, 2)), 
                np.zeros((0, 2))]
        pim3 = PersImage(pixels = (10, 10))
        res3 = pim3.transform(dgm3)
        assert np.all(res3[0] == np.zeros((10, 10)))
        assert np.all(res3[1] == np.zeros((10, 10)))


class TestWeighting:
    def test_zero_on_xaxis(self):
        pim = PersImage()

        wf = pim.weighting()

        assert wf([1, 0]) == 0
        assert wf([100, 0]) == 0
        assert wf([99, 1.4]) == 1.4

    def test_scales(self):
        pim = PersImage()

        wf = pim.weighting(np.array([[0, 1], [1, 2], [3, 4]]))

        assert wf([1, 0]) == 0
        assert wf([1, 4]) == 1
        assert wf([1, 2]) == .5


class TestKernels:
    def test_kernel_mean(self):
        pim = PersImage()
        kf = pim.kernel(2)

        data = np.array([[0, 0]])
        assert kf(np.array([[0, 0]]), [0, 0]) >= kf(
            np.array([[1, 1]]), [0, 0]
        ), "decreasing away"
        assert kf(np.array([[0, 0]]), [1, 1]) == kf(
            np.array([[1, 1]]), [0, 0]
        ), "symmetric"


class TestTransforms:
    def test_lists_of_lists(self):
        pim = PersImage(pixels=(3, 3))
        diagram = [[0, 1], [1, 1], [3, 5]]
        img = pim.transform(diagram)

        assert img.shape == (3, 3)

    def test_n_pixels(self):
        pim = PersImage(pixels=(3, 3))
        diagram = np.array([[0, 1], [1, 1], [3, 5]])
        img = pim.transform(diagram)

        assert img.shape == (3, 3)

    def test_multiple_diagrams(self):
        pim = PersImage(pixels=(3, 3))

        diagram1 = np.array([[0, 1], [1, 1], [3, 5]])
        diagram2 = np.array([[0, 1], [1, 1], [3, 6]])
        imgs = pim.transform([diagram1, diagram2])

        assert len(imgs) == 2
        assert imgs[0].shape == imgs[1].shape



# -*- coding: utf-8 -*-
"""
    Implementation of the modified Gromov-Hausdorff (mGH) distance
    between compact metric spaces induced by unweighted graphs based on
    their shortest path length. The mGH distance was first defined in
    "Some properties of Gromov–Hausdorff distances" by F. Mémoli
    (Discrete & Computational Geometry, 2012).

    Author: Vladyslav Oles

    ===================================================================

    Usage examples:

    1) Estimating the mGH distance between 4-clique and single-vertex
    graph from their adjacency matrices. Note that it suffices to fill
    only the upper triangle of an adjacency matrix.

    >>> A_G = [[0, 1, 1, 1], [0, 0, 1, 1], [0, 0, 0, 1], [0, 0, 0, 0]]
    >>> A_H = [[0]]
    >>> lb, ub = gromov_hausdorff(A_G, A_H)
    >>> lb, ub
    (0.5, 0.5)

    2) Estimating the mGH distance between cycle graphs of length 2 and
    5 from their adjacency matrices. Note that the adjacency matrices
    can be given in both dense and sparse SciPy formats.

    >>> A_I = np.array([[0, 1], [0, 0]])
    >>> A_J = sps.csr_matrix(([1] * 5, ([0, 0, 1, 2, 3], [1, 4, 2, 3, 4])), shape=(5, 5))
    >>> lb, ub = gromov_hausdorff(A_I, A_J)
    >>> print(lb, ub)
    (0.5, 1.0)

    3) Estimating all pairwise mGH distances between multiple graphs
    from their adjacency matrices as an iterable.

    >>> As = [A_G, A_H, A_I, A_J]
    >>> LB, UB = gromov_hausdorff(As)
    >>> LB
    [[0.  0.5 0.5 0.5]
     [0.5 0.  0.5 1. ]
     [0.5 0.5 0.  0.5]
     [0.5 1.  0.5 0. ]]
    >>> UB
    [[0.  0.5 0.5 0.5]
     [0.5 0.  0.5 1. ]
     [0.5 0.5 0.  1. ]
     [0.5 1.  1.  0. ]]

    ===================================================================

    Notations:

    |X| denotes the number of elements in set X.

    X → Y denotes the set of all mappings of set X into set Y.

    V(G) denotes vertex set of graph G.

    mGH(X, Y) denotes the modified Gromov-Hausdorff distance between
    compact metric spaces X and Y.

    ===================================================================

    Glossary:

    Distance matrix of metric space X is a |X|×|X| matrix whose
    (i, j)-th entry holds the distance between i-th and j-th points of
    X. By the properties of a metric, distance matrices are symmetric
    and non-negative, their diagonal entries are 0 and off-diagonal
    entries are positive.

    Curvature is a generalization of distance matrix that allows
    repetitions in the underlying points of a metric space. Curvature
    of an n-tuple of points from metric space X is an n×n matrix whose
    (i, j)-th entry holds the distance between the points from i-th and
    j-th positions of the tuple. Since these points need not be
    distinct, the off-diagonal entries of a curvature can equal 0.

    n-th curvature set of metric space X is the set of all curvatures
    of X that are of size n×n.

    d-bounded curvature for some d > 0 is a curvature whose
    off-diagonal entries are all ≥ d.

    Positive-bounded curvature is a curvature whose off-diagonal
    entries are all positive, i.e. the points in the underlying tuple
    are distinct. Equivalently, positive-bounded curvatures are
    distance matrices on the subsets of a metric space.
"""
import numpy as np
import warnings
import scipy.sparse as sps
from scipy.sparse.csgraph import shortest_path, connected_components


__all__ = ["gromov_hausdorff"]


# To sample √|X| * log (|X| + 1) mappings from X → Y by default.
DEFAULT_MAPPING_SAMPLE_SIZE_ORDER = np.array([.5, 1])


def gromov_hausdorff(
        A_G, A_H=None, mapping_sample_size_order=DEFAULT_MAPPING_SAMPLE_SIZE_ORDER):
    """
    Estimate the mGH distance between simple unweighted graphs,
    represented as compact metric spaces based on their shortest
    path lengths.

    Parameters
    -----------
    A_G: np.array (|V(G)|×|V(G)|)
        (Sparse) adjacency matrix of graph G, or an iterable of
        adjacency matrices if A_H=None.
    A_H: np.array (|V(H)|×|V(H)|)
        (Sparse) adjacency matrix of graph H, or None.
    mapping_sample_size_order: np.array (2)
        Parameter that regulates the number of mappings to sample when
        tightening upper bound of the mGH distance.

    Returns
    --------
    lb: float
        Lower bound of the mGH distance, or a square matrix holding
        lower bounds of pairwise mGH distances if A_H=None.
    ub: float
        Upper bound of the mGH distance, or a square matrix holding
        upper bounds of pairwise mGH distances if A_H=None.
    """
    # Form iterable with adjacency matrices.
    if A_H is None:
        if len(A_G) < 2:
            raise ValueError("'estimate_between_unweighted_graphs' needs at least"
                             "2 graphs to discriminate")
        As = A_G
    else:
        As = (A_G, A_H)

    n = len(As)
    # Find lower and upper bounds of each pairwise mGH distance between
    # the graphs.
    lbs = np.zeros((n, n))
    ubs = np.zeros((n, n))
    for i in range(n):
        for j in range(i + 1, n):
            # Transform adjacency matrices of a pair of graphs to
            # distance matrices.
            D_X = make_distance_matrix_from_adjacency_matrix(As[i])
            D_Y = make_distance_matrix_from_adjacency_matrix(As[j])
            # Find lower and upper bounds of the mGH distance between
            # the pair of graphs.
            lbs[i, j], ubs[i, j] = estimate(
                D_X, D_Y, mapping_sample_size_order=mapping_sample_size_order)

    if A_H is None:
        # Symmetrize matrices with lower and upper bounds of pairwise
        # mGH distances between the graphs.
        lower_triangle_indices = np.tril_indices(n, -1)
        lbs[lower_triangle_indices] = lbs.T[lower_triangle_indices]
        ubs[lower_triangle_indices] = ubs.T[lower_triangle_indices]

        return lbs, ubs
    else:
        return lbs[0, 1], ubs[0, 1]


def make_distance_matrix_from_adjacency_matrix(A_G):
    """
    Represent simple unweighted graph as compact metric space (with
    integer distances) based on its shortest path lengths.

    Parameters
    -----------
    A_G: np.array (|V(G)|×|V(G)|)
        (Sparse) adjacency matrix of simple unweighted graph G.

    Returns
    --------
    D_G: np.array (|V(G)|×|V(G)|)
        (Dense) distance matrix of G, represented as compact
        metric space based on its shortest path lengths.
    """
    # Convert adjacency matrix to SciPy format if needed.
    if not sps.issparse(A_G) and not isinstance(A_G, np.ndarray):
        A_G = np.asarray(A_G)

    # Compile distance matrix of the graph based on its shortest path
    # lengths.
    D_G = shortest_path(A_G, directed=False, unweighted=True)
    # Ensure compactness of metric space, represented by distance
    # matrix.
    if np.any(np.isinf(D_G)):
        warnings.warn("disconnected graph is approximated by its largest connected component")
        # Extract largest connected component of the graph.
        _, components_by_vertex = connected_components(A_G, directed=False)
        components, component_sizes = np.unique(components_by_vertex, return_counts=True)
        largest_component = components[np.argmax(component_sizes)]
        D_G = D_G[components_by_vertex == largest_component]

    # Cast distance matrix to optimal integer type.
    D_G = cast_distance_matrix_to_optimal_integer_type(D_G)

    return D_G


def cast_distance_matrix_to_optimal_integer_type(D_X):
    """
    Cast distance matrix to smallest signed integer type, sufficient
    to hold all its distances.

    Parameters
    -----------
    D_X: np.array (|X|×|X|)
        Distance matrix of a compact metric space X with integer
        distances.

    Returns
    --------
    D: np.array (|X|×|X|)
        Distance matrix of the metric space, cast to optimal type.
    """
    max_distance = np.max(D_X)
    # Type is signed integer to allow subtractions.
    optimal_int_type = determine_optimal_int_type(max_distance)
    D_X = D_X.astype(optimal_int_type)

    return D_X


def determine_optimal_int_type(value):
    """
    Determine smallest signed integer type sufficient to hold a value.

    Parameters
    -----------
    value: non-negative integer

    Returns
    --------
    optimal_int_type: np.dtype
        Optimal signed integer type to hold the value.
    """
    feasible_int_types = (int_type for int_type in [np.int8, np.int16, np.int32, np.int64]
                          if value <= np.iinfo(int_type).max)
    try:
        optimal_int_type = next(feasible_int_types)
    except StopIteration:
        raise ValueError("value {} too large to be stored as unsigned integer")

    return optimal_int_type


def estimate(D_X, D_Y, mapping_sample_size_order=DEFAULT_MAPPING_SAMPLE_SIZE_ORDER):
    """
    For X, Y metric spaces, find lower and upper bounds of mGH(X, Y).

    Parameters
    ----------
    D_X: np.array (|X|×|X|)
        Integer distance matrix of X.
    D_Y: np.array (|Y|×|Y|)
        Integer distance matrix of Y.
    mapping_sample_size_order: np.array (2)
        Parameter that regulates the number of mappings to sample when
        tightening upper bound of mGH(X, Y).

    Returns
    --------
    lb: float
        Lower bound of mGH(X, Y).
    ub: float
        Upper bound of mGH(X, Y).
    """
    # Ensure distance matrices are of integer type.
    if not np.issubdtype(D_X.dtype, np.integer) or not np.issubdtype(D_Y.dtype, np.integer):
        raise ValueError("non-integer metrics are not yet supported")
    # Cast distance matrices to signed integer type to allow
    # subtractions.
    if np.issubdtype(D_X.dtype, np.uint):
        D_X = cast_distance_matrix_to_optimal_integer_type(D_X)
    if np.issubdtype(D_Y.dtype, np.uint):
        D_Y = cast_distance_matrix_to_optimal_integer_type(D_Y)
    # Estimate mGH(X, Y).
    double_lb = find_lb(D_X, D_Y)
    double_ub = find_ub(
        D_X, D_Y, mapping_sample_size_order=mapping_sample_size_order, double_lb=double_lb)

    return 0.5 * double_lb, 0.5 * double_ub


def find_lb(D_X, D_Y):
    """
    For X, Y metric spaces, find lower bound of mGH(X, Y).

    Parameters
    ----------
    D_X: np.array (|X|×|X|)
        Integer distance matrix of X.
    D_Y: np.array (|Y|×|Y|)
        Integer distance matrix of Y.

    Returns
    --------
    double_lb: float
        Lower bound of 2*mGH(X, Y).
    """
    diam_X = np.max(D_X)
    diam_Y = np.max(D_Y)
    max_distance = max(diam_X, diam_Y)
    # Obtain trivial lower bound of 2*mGH(X, Y) from
    # 1) mGH(X, Y) ≥ 0.5*|diam X - diam Y|;
    # 2) mGH(X, Y) = 0 if and only if X and Y are isometric.
    trivial_double_lb = max(abs(diam_X - diam_Y), int(len(D_X) != len(D_Y)))
    # Initialize lower bound of 2*mGH(X, Y).
    double_lb = trivial_double_lb
    # Try tightening the lower bound.
    d = max_distance
    while d > double_lb:
        # Try proving 2*mGH(X, Y) ≥ d using d-bounded curvatures of
        # X and Y of size 3×3 or larger. 2×2 curvatures are already
        # accounted for in trivial lower bound.

        if d <= diam_X:
            K = find_largest_size_bounded_curvature(D_X, diam_X, d)
            if len(K) > 2 and confirm_lb_using_bounded_curvature(d, K, D_Y, max_distance):
                double_lb = d
        if d > double_lb and d <= diam_Y:
            L = find_largest_size_bounded_curvature(D_Y, diam_Y, d)
            if len(L) > 2 and confirm_lb_using_bounded_curvature(d, L, D_X, max_distance):
                double_lb = d

        d -= 1

    return double_lb


def find_largest_size_bounded_curvature(D_X, diam_X, d):
    """
    Find a largest-size d-bounded curvature of metric space X.

    Parameters
    ----------
    D_X: np.array (|X|×|X|)
        Integer distance matrix of X.
    diam_X: int
        Largest distance in X.

    Returns
    --------
    K: np.array (N×N)
        d-bounded curvature of X of largest size; N ≤ |K|.
    """
    # Initialize curvature K with entire distance matrix.
    K = D_X
    while np.any(K[np.triu_indices_from(K, 1)] < d):
        # Pick a row (and column) with highest number of off-diagonal
        # distances < d, then with smallest sum of off-diagonal
        # distances ≥ d.
        K_rows_sortkeys = -np.sum(K < d, axis=0) * (len(K) * diam_X) + \
                      np.sum(np.ma.masked_less(K, d), axis=0).data
        row_to_remove = np.argmin(K_rows_sortkeys)
        # Remove the point from K.
        K = np.delete(K, row_to_remove, axis=0)
        K = np.delete(K, row_to_remove, axis=1)

    return K


def confirm_lb_using_bounded_curvature(d, K, D_Y, max_distance):
    """
    For X, Y metric spaces, try to confirm 2*mGH(X, Y) ≥ d
    using K, a d-bounded curvature of X.

    Parameters
    ----------
    d: int
        Lower bound candidate for 2*mGH(X, Y).
    K: np.array (N×N)
        d-bounded curvature of X; N ≥ 3.
    D_Y: np.array (|Y|×|Y|)
        Integer distance matrix of Y.
    max_distance: int
        Largest distance in X and Y.

    Returns
    --------
    lb_is_confirmed: bool
        Whether confirmed that 2*mGH(X, Y) ≥ d.
    """
    # If K exceeds D_Y in size, the Hausdorff distance between the n-th
    # curvature sets of X and Y is ≥ d, entailing 2*mGH(X, Y) ≥ d.
    lb_is_confirmed = len(D_Y) < len(K) or confirm_lb_using_bounded_curvature_principal_subrows(
        d, K, D_Y, max_distance)

    return lb_is_confirmed


def confirm_lb_using_bounded_curvature_principal_subrows(d, K, D_Y, max_distance):
    """
    For X, Y metric spaces, and K, a d-bounded curvature of X, try to
    confirm 2*mGH(X, Y) ≥ d using selected rows of selected principal
    submatrices of K.

    Parameters
    ----------
    d: int
        Lower bound candidate for 2*mGH(X, Y).
    K: np.array (N×N)
        d-bounded curvature of X; N ≥ 3.
    D_Y: np.array (|Y|×|Y|)
        Integer distance matrix of Y; |Y| ≥ N.
    max_distance: int
        Largest distance in X and Y.

    Returns
    --------
    lb_is_confirmed: bool
        Whether confirmed that 2*mGH(X, Y) ≥ d.
    """
    # Represent rows of D_Y as distance distributions.
    D_Y_rows_distance_distributions = \
        represent_distance_matrix_rows_as_distributions(D_Y, max_distance)

    # For each m ≥ 3, select those rows of m×m principal
    # submatrices of K whose entries are the largest, represent them as
    # distance distributions, and try using them to confirm that the
    # Hausdorff distance between m-th curvature sets of X and Y is ≥ d,
    # entailing 2*mGH(X, Y) ≥ d. The case of m = 2 is disregarded as
    # 2×2 curvatures are already accounted for in trivial lower bound.
    K_principal_subrows_distance_distributions = find_unique_maximal_distributions(
        represent_distance_matrix_rows_as_distributions(K, max_distance))
    lb_is_confirmed = confirm_distance_between_curvature_sets_lb(
        K_principal_subrows_distance_distributions, D_Y_rows_distance_distributions, d)
    m = len(K) - 1
    while not lb_is_confirmed and m > 2:
        K_principal_subrows_distance_distributions = find_unique_maximal_distributions(
            remove_smallest_entry_from_vectors(
                K_principal_subrows_distance_distributions))
        lb_is_confirmed = confirm_distance_between_curvature_sets_lb(
            K_principal_subrows_distance_distributions, D_Y_rows_distance_distributions, d)

        m -= 1

    return lb_is_confirmed


def represent_distance_matrix_rows_as_distributions(D, max_distance):
    """
    Represent each row of a distance matrix by frequency distribution
    of distances in it. The only distance 0 in each row is omitted.

    Parameters
    ----------
    D: np.array (N×N)
        Integer distance matrix.
    max_distance: int
        Max distance of the distribution, upper bound of entries in D.

    Returns
    --------
    D_rows_distributons: np.array (N×max_distance)
        Each row holds frequencies of each distance from 1 to
        max_distance in the corresponding row of D:
        (i, j)-th entry holds frequency of distance (max_distance - j)
        in the i-th row of D.
    """
    # Add imaginary part to distinguish identical distances from
    # different rows of D.
    unique_distances, distance_frequencies = np.unique(
        D + 1j * np.arange(len(D))[:, None], return_counts=True)
    # Type is signed integer to allow subtractions.
    optimal_int_type = determine_optimal_int_type(len(D))
    D_rows_distributons = np.zeros((len(D), max_distance + 1), dtype=optimal_int_type)
    # Construct index pairs for distance frequencies, so that the
    # frequencies of larger distances appear on the left.
    distance_frequencies_index_pairs = \
        (np.imag(unique_distances).astype(optimal_int_type),
         max_distance - np.real(unique_distances).astype(max_distance.dtype))
    # Fill frequency distributions of D rows.
    D_rows_distributons[distance_frequencies_index_pairs] = distance_frequencies
    # Remove (unit) frequency of distance 0 from each row.
    D_rows_distributons = D_rows_distributons[:, :-1]

    return D_rows_distributons


def find_unique_maximal_distributions(distributions):
    """
    Given some vectors represented as frequency distributions of their
    entries, find unique maximal vectors under the following partial
    order: for r, s vectors, r < s if and only if there exists
    bijection between the individual entries of r and s, under which
    each entry in r is smaller than the corresponding entry in s.

    Parameters
    ----------
    distributions: np.array (N×max_distance)
        Frequency distributions of N positive vectors of the same size,
        i.e. the row sums are all equal.

    Returns
    --------
    unique_maximal_distributions: np.array (M, max_distance)
        Unique frequency distributions of the maximal vectors; M ≤ N.
    """
    pairwise_distribution_differences = \
        np.cumsum(distributions - distributions[:, None, :], axis=2)
    pairwise_distribution_less_thans = np.logical_and(
        np.all(pairwise_distribution_differences >= 0, axis=2),
        np.any(pairwise_distribution_differences > 0, axis=2))
    distributions_are_maximal = ~np.any(pairwise_distribution_less_thans, axis=1)
    # `np.unique` replaced for compatibility with Python 3.4/NumPy 1.12.
    # unique_maximal_distributions = np.unique(distributions[distributions_are_maximal], axis=0)
    unique_maximal_distributions = np.vstack(
        {tuple(distribution) for distribution in distributions[distributions_are_maximal]})

    return unique_maximal_distributions


def confirm_distance_between_curvature_sets_lb(rs_distributions, D_Y_rows_distributions, d):
    """
    For X, Y metric spaces, try to confirm that the Hausdorff distance
    between the m-th curvature sets of X and Y is ≥ d, using a row of
    some m×m d-bounded curvature of X.

    Parameters
    ----------
    rs_distributions: np.array (m×max_distance)
        Frequency distributions of the rows of some m×m d-bounded
        curvature of X; m ≥ 3.
    D_Y_rows_distributions: np.array (|Y|×max_distance)
        Frequency distributions of the rows of distance matrix of Y;
        |Y| ≥ m.
    d: int
        Lower bound candidate for the Hausdorff distance between the
        m-th curvature sets of X and Y.

    Returns
    --------
    distance_between_curvature_sets_lb_is_confirmed: bool
        Whether confirmed that the Hausdorff distance between the
        m-th curvature sets of X and Y is ≥ d.
    """
    # For each r, k-th row of an m×m d-bounded curvature of X for some
    # k ∈ 1,...,m, check if the l∞-distance from r to the set of k-th
    # rows of positive-bounded m×m curvatures of Y is ≥ d, then the
    # Hausdorff distance between m-th curvature sets of X and Y is ≥ d.
    distance_between_curvature_sets_lb_is_confirmed = False
    i = 0
    while not distance_between_curvature_sets_lb_is_confirmed and i < len(rs_distributions):
        distance_between_curvature_sets_lb_is_confirmed = True
        j = 0
        while distance_between_curvature_sets_lb_is_confirmed and j < len(D_Y_rows_distributions):
            distance_between_curvature_sets_lb_is_confirmed = \
                check_distance_to_permuted_principal_subrows_lb(
                    rs_distributions[i], D_Y_rows_distributions[j], d)
            j += 1

        i += 1

    return distance_between_curvature_sets_lb_is_confirmed


def remove_smallest_entry_from_vectors(distributions):
    """
    Remove smallest entry from each of the given vectors, represented
    as frequency distributions of their entries.

    Parameters
    ----------
    distributions: np.array (N×max_distance)
        Frequency distributions of N positive vectors.

    Returns
    --------
    updated_distributions: np.array (N×max_distance)
        Frequency distributions of N vectors, obtained by removing
        smallest entry from each of the input vectors.
    """
    updated_distributions = distributions.copy()
    # Find smallest entry in each distribution.
    max_distance = distributions.shape[1]
    smallest_entry_indices = max_distance - 1 - np.argmin(np.fliplr(distributions) == 0, axis=1)
    # Decrease frequency of smallest entry by 1 in each distribution.
    updated_distributions[np.arange(len(distributions)), smallest_entry_indices] -= 1

    return updated_distributions


def check_distance_to_permuted_principal_subrows_lb(r_distribution, s_distribution, d):
    """
    For vectors r of size m and s, a row of some n×n distance matrix D,
    check if l∞-distance from r to the set P, comprised of permutations
    of those subvectors of s that are a row of some m×m principal
    submatrix of D, is ≥ d.

    Parameters
    ----------
    r_distribution: np.array (max_distance)
        Frequency distribution of r, positive integer vector.
    s_distribution: np.array (max_distance)
        Frequency distribution of s, positive integer vector no smaller
        in length than r;
    d: int
        Lower bound candidate for l∞-distance from r to P; d > 0.

    Returns
    --------
    d_is_distance_to_permuted_principal_subrows_lb: bool
        Whether l∞-distance from r to P is ≥ d.
    """
    def next_i_and_j(min_i, min_j):
        # Find reversed r distribution index of smallest r entries yet
        # to be assigned. Then find index in reversed s distribution of
        # smallest s entries to which the r entries can be assigned to.
        try:
            i = next(i for i in range(min_i, len(reversed_r_distribution))
                     if reversed_r_distribution[i] > 0)
        except StopIteration:
            # All r entries are assigned.
            i = None
            j = min_j
        else:
            j = next_j(i, max(i - (d - 1), min_j))

        return i, j

    def next_j(i, min_j):
        # Find reversed s distribution index of smallest s entries to
        # which r entries, corresponding to a given reversed r
        # distribution index, can be assigned to.
        try:
            j = next(j for j in range(min_j, min(i + (d - 1),
                                                 len(reversed_s_distribution) - 1) + 1)
                     if reversed_s_distribution[j] > 0)
        except StopIteration:
            # No s entries left to assign the particular r entries to.
            j = None

        return j

    # Copy to allow modifications and stay pure; reverse for the
    # frequencies of smaller entries to come first, to be compatible
    # even with distributions of different lengths.
    reversed_r_distribution = list(r_distribution[::-1])
    reversed_s_distribution = list(s_distribution[::-1])
    # Injectively assign r entries to s entries if their difference
    # is < d, going from smallest to largest entries in both r and s,
    # until all r entries are assigned or such assignment proves
    # unfeasible.
    i, j = next_i_and_j(0, 0)
    while i is not None and j is not None:
        if reversed_r_distribution[i] <= reversed_s_distribution[j]:
            reversed_s_distribution[j] -= reversed_r_distribution[i]
            reversed_r_distribution[i] = 0
            i, j = next_i_and_j(i, j)
        else:
            reversed_r_distribution[i] -= reversed_s_distribution[j]
            reversed_s_distribution[j] = 0
            j = next_j(i, j)

    # The assignment is feasible if and only if |r - p| < d for some
    # p ∈ P, and therefore infeasible if and only if l∞-distance from
    # r to P is ≥ d.
    d_is_distance_to_permuted_principal_subrows_lb = (j is None)

    return d_is_distance_to_permuted_principal_subrows_lb


def find_ub(D_X, D_Y, mapping_sample_size_order=DEFAULT_MAPPING_SAMPLE_SIZE_ORDER, double_lb=0):
    """
    For X, Y metric spaces, find upper bound of mGH(X, Y).

    Parameters
    ----------
    D_X: np.array (|X|×|X|)
        Integer distance matrix of X.
    D_Y: np.array (|Y|×|Y|)
        Integer distance matrix of Y.
    mapping_sample_size_order: np.array (2)
        Parameter that regulates the number of mappings to sample when
        tightening the upper bound.
    double_lb: float
        Lower bound of 2*mGH(X, Y).

    Returns
    --------
    double_ub: float
        Upper bound of 2*mGH(X, Y).
    """
    # Find upper bound of smallest distortion of a mapping in X → Y.
    ub_of_X_to_Y_min_distortion = find_ub_of_min_distortion(
        D_X, D_Y, mapping_sample_size_order=mapping_sample_size_order, goal_distortion=double_lb)
    # Find upper bound of smallest distortion of a mapping in Y → X.
    ub_of_Y_to_X_min_distortion = find_ub_of_min_distortion(
        D_Y, D_X, mapping_sample_size_order=mapping_sample_size_order,
        goal_distortion=ub_of_X_to_Y_min_distortion)

    return max(ub_of_X_to_Y_min_distortion, ub_of_Y_to_X_min_distortion)


def find_ub_of_min_distortion(D_X, D_Y,
                              mapping_sample_size_order=DEFAULT_MAPPING_SAMPLE_SIZE_ORDER,
                              goal_distortion=0):
    """
    For X, Y metric spaces, finds upper bound of smallest distortion of
    a mapping in X → Y by heuristically constructing some mappings and
    choosing the smallest distortion in the sample.

    Parameters
    ----------
    D_X: np.array (|X|×|X|)
        Integer distance matrix of X.
    D_Y: np.array (|Y|×|Y|)
        Integer distance matrix of Y.
    mapping_sample_size_order: np.array (2)
        Exponents of |X| and log (|X|+1) in their product that defines
        how many mappings from X → Y to sample.
    goal_distortion: float
        No need to look for distortion smaller than this.

    Returns
    --------
    ub_of_min_distortion: float
        Upper bound of smallest distortion of a mapping in X → Y.
    """
    # Compute the numper of mappings to sample.
    n_mappings_to_sample = int(np.ceil(np.prod(
        np.array([len(D_X), np.log(len(D_X) + 1)])**mapping_sample_size_order)))
    # Construct each mapping in X → Y in |X| steps by choosing the image
    # of π(i)-th point in X at i-th step, where π is randomly sampled
    # |X|-permutation. Image of each point is chosen to minimize the
    # intermediate distortion at each step.
    permutations_generator = (np.random.permutation(len(D_X)) for _ in range(n_mappings_to_sample))
    ub_of_min_distortion = np.inf
    goal_distortion_is_matched = False
    all_sampled_permutations_are_tried = False
    pi = next(permutations_generator)
    while not goal_distortion_is_matched and not all_sampled_permutations_are_tried:
        mapped_xs_images, distortion = construct_mapping(D_X, D_Y, pi)
        ub_of_min_distortion = min(distortion, ub_of_min_distortion)
        if ub_of_min_distortion <= goal_distortion:
            goal_distortion_is_matched = True

        try:
            pi = next(permutations_generator)
        except StopIteration:
            all_sampled_permutations_are_tried = True

    return ub_of_min_distortion


def construct_mapping(D_X, D_Y, pi):
    """
    # Construct a mapping from X → Y in |X| steps by choosing the image
    # of π(i)-th point in X at i-th step, for some |X|-permutation π.
    Image of each point is chosen to minimize the intermediate
    distortion at the corresponding step.

    D_X: np.array (|X|×|X|)
        Integer distance matrix of X.
    D_Y: np.array (|Y|×|Y|)
        Integer distance matrix of Y.
    pi: np.array (|X|)
        |X|-permutation specifying the order in which the points in X
        are mapped.
    Returns
    --------
    mapped_xs_images: list
        image of the constructed mapping.
    distortion: float
        distortion of the constructed mapping.
    """
    # Map π(1)-th point in X to a random point in Y, due to the
    # lack of better criterion.
    mapped_xs = [pi[0]]
    mapped_xs_images = [np.random.choice(len(D_Y))]
    distortion = 0
    # Map π(i)-th point in X for each i = 2,...,|X|.
    for x in pi[1:]:
        # Choose point in Y that minimizes the distortion after
        # mapping π(i)-th point in X to it.
        bottlenecks_from_mapping_x = np.max(
            np.abs(D_X[x, mapped_xs] - D_Y[:, mapped_xs_images]), axis=1)
        y = np.argmin(bottlenecks_from_mapping_x)
        # Map π(i)-th point in X to the chosen point in Y.
        mapped_xs.append(x)
        mapped_xs_images.append(y)
        distortion = max(bottlenecks_from_mapping_x[y], distortion)
        
    return mapped_xs_images, distortion




import numpy as np
import scipy.sparse as sps

import pytest

from persim import bottleneck, wasserstein
from persim import sliced_wasserstein
from persim import heat
from persim import gromov_hausdorff, gromov_hausdorff


class TestBottleneck:
    def test_single(self):
        d = bottleneck(
            np.array([[0.5, 1]]),
            np.array([[0.5, 1.1]])
        )

        # These are very loose bounds
        assert d == pytest.approx(0.1, 0.001)

    def test_some(self):
        d = bottleneck(
            np.array([
                [0.5, 1],
                [0.6, 1.1]
            ]),
            np.array([
                [0.5, 1.1],
                [0.6, 1.3]
            ])
        )

        # These are very loose bounds
        assert d == pytest.approx(0.2, 0.001)

    def test_diagonal(self):
        d = bottleneck(
            np.array([
                [10.5, 10.5],
                [10.6, 10.5],
                [10.3, 10.3]
            ]),
            np.array([
                [0.5, 1.0],
                [0.6, 1.2],
                [0.3, 0.7]
            ])
        )

        # I expect this to be 0.6
        assert d == pytest.approx(0.3, 0.001)

    def test_different_size(self):
        d = bottleneck(
            np.array([
                [0.5, 1],
                [0.6, 1.1]
            ]),
            np.array([
                [0.5, 1.1]
            ])
        )

        # These are very loose bounds
        assert d == pytest.approx(0.1, 0.001)

    def test_matching(self):
        dgm1 = np.array([
            [0.5, 1],
            [0.6, 1.1]
        ])
        dgm2 = np.array([
            [0.5, 1.1],
            [0.6, 1.1],
            [0.8, 1.1],
            [1.0, 1.1],
        ])

        d, (m, D) = bottleneck(
            dgm1, dgm2,
            matching=True
        )

        # These are very loose bounds
        assert len(m) == len(dgm1) + len(dgm2)
        assert D.shape == (len(dgm1) + len(dgm2), len(dgm1) + len(dgm2))
    
    def test_matching_to_self(self):
        # Matching a diagram to itself should yield 0
        pd = np.array([[0.        , 1.71858561],
                      [0.        , 1.74160683],
                      [0.        , 2.43430877],
                      [0.        , 2.56949258],
                      [0.        , np.inf]])
        dist = bottleneck(pd, pd)
        assert dist == 0

class TestWasserstein:
    def test_single(self):
        d = wasserstein(
            np.array([[0.5, 1]]),
            np.array([[0.5, 1.1]])
        )

        # These are very loose bounds
        assert d == pytest.approx(0.1, 0.001)

    def test_some(self):
        d = wasserstein(
            np.array([
                [0.6, 1.1],
                [0.5, 1]
            ]),
            np.array([
                [0.5, 1.1],
                [0.6, 1.3]
            ])
        )

        # These are very loose bounds
        assert d == pytest.approx(0.3, 0.001)

    def test_matching_to_self(self):
        # Matching a diagram to itself should yield 0
        pd = np.array([[0.        , 1.71858561],
                      [0.        , 1.74160683],
                      [0.        , 2.43430877],
                      [0.        , 2.56949258]])
        dist = wasserstein(pd, pd)
        assert dist == 0


class TestSliced:
    def test_single(self):
        d = sliced_wasserstein(
            np.array([[0.5, 1]]),
            np.array([[0.5, 1.1]])
        )

        # These are very loose bounds
        assert d == pytest.approx(0.1, 0.01)

    def test_some(self):
        d = sliced_wasserstein(
            np.array([
                [0.5, 1],
                [0.6, 1.1]
            ]),
            np.array([
                [0.5, 1.1],
                [0.6, 1.2]
            ])
        )

        # These are very loose bounds
        assert d == pytest.approx(0.19, 0.02)

    def test_different_size(self):
        d = sliced_wasserstein(
            np.array([
                [0.5, 1],
                [0.6, 1.1]
            ]),
            np.array([
                [0.6, 1.2]
            ])
        )

        # These are very loose bounds
        assert d == pytest.approx(0.314, 0.1)


class TestHeat:
    def test_compare(self):
        """ lets at least be sure that large distances are captured """
        d1 = heat(
            np.array([[0.5, 1]]),
            np.array([[0.5, 1.1]])
        )
        d2 = heat(
            np.array([[0.5, 1]]),
            np.array([[0.5, 1.5]])
        )

        # These are very loose bounds
        assert d1 < d2


class TestModifiedGromovHausdorff:
    def test_single_point(self):
        A_G = sps.csr_matrix(
            ([1]*4, ([0, 0, 1, 2], [1, 3, 2, 3])), shape=(4, 4))
        A_H = sps.csr_matrix(([], ([], [])), shape=(1, 1))
        lb, ub = gromov_hausdorff(A_G, A_H)

        assert lb == 1
        assert ub == 1

    def test_isomorphic(self):
        A_G = sps.csr_matrix(
            ([1]*6, ([0, 0, 0, 1, 1, 2], [1, 2, 3, 2, 3, 3])), shape=(4, 4))
        A_H = sps.csr_matrix(
            ([1]*6, ([0, 0, 0, 1, 1, 2], [1, 2, 3, 2, 3, 3])), shape=(4, 4))
        lb, ub = gromov_hausdorff(A_G, A_H)

        assert lb == 0
        assert ub == 0

    def test_cliques(self):
        A_G = sps.csr_matrix(([1], ([0], [1])), shape=(2, 2))
        A_H = sps.csr_matrix(
            ([1]*6, ([0, 0, 0, 1, 1, 2], [1, 2, 3, 2, 3, 3])), shape=(4, 4))
        lb, ub = gromov_hausdorff(A_G, A_H)

        assert lb == 0.5
        assert ub == 0.5

    def test_same_size(self):
        A_G = sps.csr_matrix(
            ([1]*6, ([0, 0, 0, 1, 1, 2], [1, 2, 3, 2, 3, 3])), shape=(4, 4))
        A_H = sps.csr_matrix(
            ([1]*4, ([0, 0, 1, 2], [1, 3, 2, 3])), shape=(4, 4))
        lb, ub = gromov_hausdorff(A_G, A_H)

        assert lb == 0.5
        assert ub == 0.5

    def test_many_graphs(self):
        A_G = sps.csr_matrix(([1], ([0], [1])), shape=(2, 2))
        A_H = sps.csr_matrix(
            ([1]*6, ([0, 0, 0, 1, 1, 2], [1, 2, 3, 2, 3, 3])), shape=(4, 4))
        A_I = sps.csr_matrix(
            ([1]*4, ([0, 0, 1, 2], [1, 3, 2, 3])), shape=(4, 4))
        lbs, ubs = gromov_hausdorff([A_G, A_H, A_I])

        np.testing.assert_array_equal(lbs, np.array(
            [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]))
        np.testing.assert_array_equal(ubs, np.array(
            [[0, 0.5, 0.5], [0.5, 0, 0.5], [0.5, 0.5, 0]]))



import numpy as np
import matplotlib.pyplot as plt

__all__ = ["plot_diagrams", "bottleneck_matching", "wasserstein_matching"]


def plot_diagrams(
    diagrams,
    plot_only=None,
    title=None,
    xy_range=None,
    labels=None,
    colormap="default",
    size=20,
    ax_color=np.array([0.0, 0.0, 0.0]),
    diagonal=True,
    lifetime=False,
    legend=True,
    show=False,
    ax=None
):
    """A helper function to plot persistence diagrams. 

    Parameters
    ----------
    diagrams: ndarray (n_pairs, 2) or list of diagrams
        A diagram or list of diagrams. If diagram is a list of diagrams, 
        then plot all on the same plot using different colors.
    plot_only: list of numeric
        If specified, an array of only the diagrams that should be plotted.
    title: string, default is None
        If title is defined, add it as title of the plot.
    xy_range: list of numeric [xmin, xmax, ymin, ymax]
        User provided range of axes. This is useful for comparing 
        multiple persistence diagrams.
    labels: string or list of strings
        Legend labels for each diagram. 
        If none are specified, we use H_0, H_1, H_2,... by default.
    colormap: string, default is 'default'
        Any of matplotlib color palettes. 
        Some options are 'default', 'seaborn', 'sequential'. 
        See all available styles with

        .. code:: python

            import matplotlib as mpl
            print(mpl.styles.available)

    size: numeric, default is 20
        Pixel size of each point plotted.
    ax_color: any valid matplotlib color type. 
        See [https://matplotlib.org/api/colors_api.html](https://matplotlib.org/api/colors_api.html) for complete API.
    diagonal: bool, default is True
        Plot the diagonal x=y line.
    lifetime: bool, default is False. If True, diagonal is turned to False.
        Plot life time of each point instead of birth and death. 
        Essentially, visualize (x, y-x).
    legend: bool, default is True
        If true, show the legend.
    show: bool, default is False
        Call plt.show() after plotting. If you are using self.plot() as part 
        of a subplot, set show=False and call plt.show() only once at the end.
    """

    ax = ax or plt.gca()
    plt.style.use(colormap)

    xlabel, ylabel = "Birth", "Death"

    if labels is None:
        # Provide default labels for diagrams if using self.dgm_
        labels = [
            "$H_0$",
            "$H_1$",
            "$H_2$",
            "$H_3$",
            "$H_4$",
            "$H_5$",
            "$H_6$",
            "$H_7$",
            "$H_8$",
        ]

    if not isinstance(diagrams, list):
        # Must have diagrams as a list for processing downstream
        diagrams = [diagrams]

    if plot_only:
        diagrams = [diagrams[i] for i in plot_only]
        labels = [labels[i] for i in plot_only]

    if not isinstance(labels, list):
        labels = [labels] * len(diagrams)

    # Construct copy with proper type of each diagram
    # so we can freely edit them.
    diagrams = [dgm.astype(np.float32, copy=True) for dgm in diagrams]

    # find min and max of all visible diagrams
    concat_dgms = np.concatenate(diagrams).flatten()
    has_inf = np.any(np.isinf(concat_dgms))
    finite_dgms = concat_dgms[np.isfinite(concat_dgms)]

    # clever bounding boxes of the diagram
    if not xy_range:
        # define bounds of diagram
        ax_min, ax_max = np.min(finite_dgms), np.max(finite_dgms)
        x_r = ax_max - ax_min

        # Give plot a nice buffer on all sides.
        # ax_range=0 when only one point,
        buffer = 1 if xy_range == 0 else x_r / 5

        x_down = ax_min - buffer / 2
        x_up = ax_max + buffer

        y_down, y_up = x_down, x_up
    else:
        x_down, x_up, y_down, y_up = xy_range

    yr = y_up - y_down

    if lifetime:

        # Don't plot landscape and diagonal at the same time.
        diagonal = False

        # reset y axis so it doesn't go much below zero
        y_down = -yr * 0.05
        y_up = y_down + yr

        # set custom ylabel
        ylabel = "Lifetime"

        # set diagrams to be (x, y-x)
        for dgm in diagrams:
            dgm[:, 1] -= dgm[:, 0]

        # plot horizon line
        ax.plot([x_down, x_up], [0, 0], c=ax_color)

    # Plot diagonal
    if diagonal:
        ax.plot([x_down, x_up], [x_down, x_up], "--", c=ax_color)

    # Plot inf line
    if has_inf:
        # put inf line slightly below top
        b_inf = y_down + yr * 0.95
        ax.plot([x_down, x_up], [b_inf, b_inf], "--", c="k", label=r"$\infty$")

        # convert each inf in each diagram with b_inf
        for dgm in diagrams:
            dgm[np.isinf(dgm)] = b_inf

    # Plot each diagram
    for dgm, label in zip(diagrams, labels):

        # plot persistence pairs
        ax.scatter(dgm[:, 0], dgm[:, 1], size, label=label, edgecolor="none")

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)

    ax.set_xlim([x_down, x_up])
    ax.set_ylim([y_down, y_up])
    ax.set_aspect('equal', 'box')

    if title is not None:
        ax.set_title(title)

    if legend is True:
        ax.legend(loc="lower right")

    if show is True:
        plt.show()


def bottleneck_matching(I1, I2, matchidx, D, labels=["dgm1", "dgm2"], ax=None):
    """ Visualize bottleneck matching between two diagrams

    Parameters
    ===========

    I1: array
        A diagram
    I2: array
        A diagram
    matchidx: tuples of matched indices
        if input `matching=True`, then return matching
    D: array
        cross-similarity matrix
    labels: list of strings
        names of diagrams for legend. Default = ["dgm1", "dgm2"], 
    ax: matplotlib Axis object
        For plotting on a particular axis.


    Examples
    ==========

    bn_matching, (matchidx, D) = persim.bottleneck(A_h1, B_h1, matching=True)
    persim.bottleneck_matching(A_h1, B_h1, matchidx, D)

    """

    plot_diagrams([I1, I2], labels=labels, ax=ax)
    cp = np.cos(np.pi / 4)
    sp = np.sin(np.pi / 4)
    R = np.array([[cp, -sp], [sp, cp]])
    if I1.size == 0:
        I1 = np.array([[0, 0]])
    if I2.size == 0:
        I2 = np.array([[0, 0]])
    I1Rot = I1.dot(R)
    I2Rot = I2.dot(R)
    dists = [D[i, j] for (i, j) in matchidx]
    (i, j) = matchidx[np.argmax(dists)]
    if i >= I1.shape[0] and j >= I2.shape[0]:
        return
    if i >= I1.shape[0]:
        diagElem = np.array([I2Rot[j, 0], 0])
        diagElem = diagElem.dot(R.T)
        plt.plot([I2[j, 0], diagElem[0]], [I2[j, 1], diagElem[1]], "g")
    elif j >= I2.shape[0]:
        diagElem = np.array([I1Rot[i, 0], 0])
        diagElem = diagElem.dot(R.T)
        plt.plot([I1[i, 0], diagElem[0]], [I1[i, 1], diagElem[1]], "g")
    else:
        plt.plot([I1[i, 0], I2[j, 0]], [I1[i, 1], I2[j, 1]], "g")


def wasserstein_matching(I1, I2, matchidx, palette=None, labels=["dgm1", "dgm2"], colors=None, ax=None):
    """ Visualize bottleneck matching between two diagrams

    Parameters
    ===========

    I1: array
        A diagram
    I2: array
        A diagram
    matchidx: tuples of matched indices
        if input `matching=True`, then return matching
    labels: list of strings
        names of diagrams for legend. Default = ["dgm1", "dgm2"], 
    ax: matplotlib Axis object
        For plotting on a particular axis.

    Examples
    ==========

    bn_matching, (matchidx, D) = persim.wasserstien(A_h1, B_h1, matching=True)
    persim.wasserstein_matching(A_h1, B_h1, matchidx, D)

    """

    cp = np.cos(np.pi / 4)
    sp = np.sin(np.pi / 4)
    R = np.array([[cp, -sp], [sp, cp]])
    if I1.size == 0:
        I1 = np.array([[0, 0]])
    if I2.size == 0:
        I2 = np.array([[0, 0]])
    I1Rot = I1.dot(R)
    I2Rot = I2.dot(R)
    for index in matchidx:
        (i, j) = index
        if i >= I1.shape[0] and j >= I2.shape[0]:
            continue
        if i >= I1.shape[0]:
            diagElem = np.array([I2Rot[j, 0], 0])
            diagElem = diagElem.dot(R.T)
            plt.plot([I2[j, 0], diagElem[0]], [I2[j, 1], diagElem[1]], "g")
        elif j >= I2.shape[0]:
            diagElem = np.array([I1Rot[i, 0], 0])
            diagElem = diagElem.dot(R.T)
            plt.plot([I1[i, 0], diagElem[0]], [I1[i, 1], diagElem[1]], "g")
        else:
            plt.plot([I1[i, 0], I2[j, 0]], [I1[i, 1], I2[j, 1]], "g")

    plot_diagrams([I1, I2], labels=labels, ax=ax)



from .images import *
from .sliced_wasserstein import *
from .bottleneck import *
from .wasserstein import *
from .heat import *
from .gromov_hausdorff import *
from .visuals import *

from ._version import __version__



import numpy as np
from scipy.spatial.distance import cityblock

__all__ = ["sliced_wasserstein"]

def sliced_wasserstein(PD1, PD2, M=50):
    """ Implementation of Sliced Wasserstein distance as described in 
        Sliced Wasserstein Kernel for Persistence Diagrams by Mathieu Carriere, Marco Cuturi, Steve Oudot (https://arxiv.org/abs/1706.03358)


        Parameters
        -----------
        
        PD1: np.array size (m,2)
            Persistence diagram
        PD2: np.array size (n,2)
            Persistence diagram
        M: int, default is 50
            Iterations to run approximation.

        Returns
        --------
        sw: float
            Sliced Wasserstein distance between PD1 and PD2
    """

    diag_theta = np.array(
        [np.cos(0.25 * np.pi), np.sin(0.25 * np.pi)], dtype=np.float32
    )

    l_theta1 = [np.dot(diag_theta, x) for x in PD1]
    l_theta2 = [np.dot(diag_theta, x) for x in PD2]

    if (len(l_theta1) != PD1.shape[0]) or (len(l_theta2) != PD2.shape[0]):
        raise ValueError("The projected points and origin do not match")

    PD_delta1 = [[np.sqrt(x ** 2 / 2.0)] * 2 for x in l_theta1]
    PD_delta2 = [[np.sqrt(x ** 2 / 2.0)] * 2 for x in l_theta2]

    # i have the input now to compute the sw
    sw = 0
    theta = 0.5
    step = 1.0 / M
    for i in range(M):
        l_theta = np.array(
            [np.cos(theta * np.pi), np.sin(theta * np.pi)], dtype=np.float32
        )

        V1 = [np.dot(l_theta, x) for x in PD1] + [np.dot(l_theta, x) for x in PD_delta2]

        V2 = [np.dot(l_theta, x) for x in PD2] + [np.dot(l_theta, x) for x in PD_delta1]

        sw += step * cityblock(sorted(V1), sorted(V2))
        theta += step

    return sw



#!/usr/bin/python
# -*- coding: utf-8 -*-

"""
    
    The persistent entropy has been defined in [1]. A precursor of this definition was given in [2]
    to measure how different bars of the barcode are in length.
    
    [1] M. Rucco, F. Castiglione, E. Merelli, M. Pettini, Characterisation of the
    idiotypic immune network through persistent entropy, in: Proc. Complex, 2015.
    [2] H. Chintakunta, T. Gentimis, R. Gonzalez-Diaz, M.-J. Jimenez,
    H. Krim, An entropy-based persistence barcode, Pattern Recognition
    48 (2) (2015) 391–401.
        
    Implementation of persistent entropy

    Author: Eduardo Paluzo Hidalgo (cimagroup, University of Seville)
    contact: epaluzo@us.es

"""

from __future__ import division
import numpy as np


__all__ = ["persistent_entropy"]


def persistent_entropy(
    dgms, keep_inf=False, val_inf=None, normalize=False
):
    """
    Perform the persistent entropy values of a family of persistence barcodes (or persistence diagrams).
    Assumes that the input diagrams are from a determined dimension. If the infinity bars have any meaning
    in your experiment and you want to keep them, remember to give the value you desire to val_Inf.

    Parameters
    -----------
    dgms: ndarray (n_pairs, 2) or list of diagrams   
        array or list of arrays of birth/death pairs of a persistence barcode of a determined dimension.
    keep_inf: bool, default False
        if False, the infinity bars are removed.
        if True, the infinity bars remain.
    val_inf: float, default None
        substitution value to infinity.
    normalize: bool, default False
        if False, the persistent entropy values are not normalized.
        if True, the persistent entropy values are normalized.
          
    Returns
    --------

    ps: ndarray (n_pairs,)
        array of persistent entropy values corresponding to each persistence barcode.

    """

    if isinstance(dgms, list) == False:
        dgms = [dgms]

    # Step 1: Remove infinity bars if keep_inf = False. If keep_inf = True, infinity value is substituted by val_inf.

    if keep_inf == False:
        dgms = [(dgm[dgm[:, 1] != np.inf]) for dgm in dgms]
    if keep_inf == True:
        if val_inf != None:
            dgms = [
                np.where(dgm == np.inf, val_inf, dgm)
                for dgm in dgms
            ]
        else:
            raise Exception(
                "Remember: You need to provide a value to infinity bars if you want to keep them."
            )

    # Step 2: Persistent entropy computation.
    ps = []
    for dgm in dgms:
        l = dgm[:, 1] - dgm[:, 0]
        if all(l > 0):
            L = np.sum(l)
            p = l / L
            E = -np.sum(p * np.log(p))
            if normalize == True:
                E = E / np.log(len(l))
            ps.append(E)
        else:
            raise Exception("A bar is born after dying")

    return np.array(ps)



import numpy as np

import pytest

from persim import persistent_entropy as pe

class TestPersistentEntropy:
    
    def test_one_diagram(self):
        dgm = np.array([[0,1],[0,3],[2,4]])
        p = pe.persistent_entropy(dgm)
        
        # An upper bound of persistent entropy is the logarithm of the number of bars.
        assert p < np.log(len(dgm))

    def test_diagrams(self):
        dgms = [np.array([[0,1],[0,3],[2,4]]),np.array([[2,5],[3,8]]),np.array([[0,10]])]
        p = pe.persistent_entropy(dgms)
        # An upper bound of persistent entropy is the logarithm of the number of bars.
        assert all(p < np.log(3))

    def test_diagrams_inf(self):
        dgms = [np.array([[0,1],[0,3],[2,4]]),np.array([[2,5],[3,8]]),np.array([[0,np.inf]])]
        p = pe.persistent_entropy(dgms, keep_inf = True, val_inf = 10)
        # An upper bound of persistent entropy is the logarithm of the number of bars.
        assert all(p < np.log(3))

    def test_diagram_one_bar(self):
        dgm = np.array([[-1,2]])
        p = pe.persistent_entropy(dgm)
        assert all(p==0)
        
        
        



"""
    Implementation of the "multiscale heat kernel" (CVPR 2015), 

    Author: Chris Tralie

"""

import numpy as np

__all__ = ["heat"]

def evalHeatKernel(dgm1, dgm2, sigma):
    """
    Evaluate the continuous heat-based kernel between dgm1 and dgm2 (more correct than L2 on the discretized version above but may be slower because can't exploit fast matrix multiplication when evaluating many, many kernels)
    """
    kSigma = 0
    I1 = np.array(dgm1)
    I2 = np.array(dgm2)
    for i in range(I1.shape[0]):
        p = I1[i, 0:2]
        for j in range(I2.shape[0]):
            q = I2[j, 0:2]
            qc = I2[j, 1::-1]
            kSigma += np.exp(-(np.sum((p - q) ** 2)) / (8 * sigma)) - np.exp(
                -(np.sum((p - qc) ** 2)) / (8 * sigma)
            )
    return kSigma / (8 * np.pi * sigma)


def heat(dgm1, dgm2, sigma=0.4):
    """
    Return the pseudo-metric between two diagrams based on the continuous
    heat kernel as described in "A Stable Multi-Scale Kernel for Topological Machine Learning" by Jan Reininghaus, Stefan Huber, Ulrich Bauer, and Roland Kwitt (CVPR 2015)

    Parameters
    -----------

    dgm1: np.array (m,2)
        A persistence diagram
    dgm2: np.array (n,2)
        A persistence diagram
    sigma: float
        Heat diffusion parameter (larger sigma makes blurrier)
    Returns
    --------

    dist: float
        heat kernel distance between dgm1 and dgm2

    """
    return np.sqrt(
        evalHeatKernel(dgm1, dgm1, sigma)
        + evalHeatKernel(dgm2, dgm2, sigma)
        - 2 * evalHeatKernel(dgm1, dgm2, sigma)
    )



#!/usr/bin/env python

from setuptools import setup


import re
VERSIONFILE="persim/_version.py"
verstrline = open(VERSIONFILE, "rt").read()
VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
mo = re.search(VSRE, verstrline, re.M)
if mo:
    verstr = mo.group(1)
else:
    raise RuntimeError("Unable to find version string in %s." % (VERSIONFILE,))

with open('README.md') as f:
    long_description = f.read()

setup(name='persim',
      version=verstr,
      description='Distances and representations of persistence diagrams',
      long_description=long_description,
      long_description_content_type="text/markdown",	
      author='Nathaniel Saul, Chris Tralie',
      author_email='nat@riverasaul.com, chris.tralie@gmail.com',
      url='https://persim.scikit-tda.org',
      license='MIT',
      packages=['persim'],
      include_package_data=True,
      install_requires=[
        'scikit-learn',
        'numpy',
        'matplotlib',
        'scipy',
        'hopcroftkarp',
      ],
      extras_require={ # use `pip install -e ".[testing]"``
        'testing': [
          'pytest' 
        ],
        'docs': [ # `pip install -e ".[docs]"``
          'sktda_docs_config'
        ]
      },
      python_requires='>=2.7,!=3.1,!=3.2,!=3.3',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Intended Audience :: Financial and Insurance Industry',
        'Intended Audience :: Healthcare Industry',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
      ],
      keywords='persistent homology, persistence images, persistence diagrams, topology data analysis, algebraic topology, unsupervised learning, sliced wasserstein distance, bottleneck distance'
     )



import pytest
import numpy as np

import matplotlib.pyplot as plt

import persim
from persim import plot_diagrams


"""

    Testing visualization is a little more difficult, but still necessary. An example of how to get started:
    > https://stackoverflow.com/questions/27948126/how-can-i-write-unit-tests-against-code-that-uses-matplotlib

    ```
    def test_plot_square2():
        f, ax = plt.subplots()
        x, y = [0, 1, 2], [0, 1, 2]
        plot_square(x, y)
        x_plot, y_plot = ax.lines[0].get_xydata().T
        np.testing.assert_array_equal(y_plot, np.square(y))

    ```


    Notes
    -----

    ax.get_children() gives all the pieces in the plot, very useful
    Scatter data is of type `PathCollection` and will be a child of ax.

"""


class TestPlotting:
    def test_single(self):
        """ Most just test this doesn't crash
        """
        diagram = np.array([[0, 1], [1, 1], [2, 4], [3, 5]])

        f, ax = plt.subplots()
        plot_diagrams(diagram, show=False)

        x_plot, y_plot = ax.lines[0].get_xydata().T

        assert x_plot[0] <= np.min(diagram)
        assert x_plot[1] >= np.max(diagram)

        # get PathCollection
        pathcols = [child for child in ax.get_children()
                    if child.__class__.__name__ == "PathCollection"]
        assert len(pathcols) == 1

    def test_multiple(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, show=False)

        pathcols = [child for child in ax.get_children()
                    if child.__class__.__name__ == "PathCollection"]

        assert len(pathcols) == 2
        np.testing.assert_array_equal(pathcols[0].get_offsets(), diagrams[0])
        np.testing.assert_array_equal(pathcols[1].get_offsets(), diagrams[1])

    def test_plot_only(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]
        f, ax = plt.subplots()
        plot_diagrams(diagrams, legend=False, show=False, plot_only=[1])

    def test_legend_true(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, legend=True, show=False)
        legend = [child for child in ax.get_children()
                  if child.__class__.__name__ == "Legend"]

        assert len(legend) == 1

    def test_legend_false(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, legend=False, show=False)
        legend = [child for child in ax.get_children()
                  if child.__class__.__name__ == "Legend"]

        assert len(legend) == 0

    def test_set_title(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, title='my title', show=False)
        assert ax.get_title() == 'my title'

        f, ax = plt.subplots()
        plot_diagrams(diagrams, show=False)
        assert ax.get_title() == ''

    def test_default_square(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, show=False)
        diagonal = ax.lines[0].get_xydata()

        assert diagonal[0, 0] == diagonal[0, 1]
        assert diagonal[1, 0] == diagonal[1, 1]

    def test_default_label(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, show=False)

        assert ax.get_ylabel() == 'Death'
        assert ax.get_xlabel() == 'Birth'

    def test_lifetime(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, lifetime=True, show=False)

        assert ax.get_ylabel() == 'Lifetime'
        assert ax.get_xlabel() == 'Birth'

        line = ax.get_lines()[0]
        np.testing.assert_array_equal(line.get_ydata(), [0, 0])

    def test_lifetime_removes_birth(self):
        diagrams = [
            np.array([[0, 1], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, lifetime=True, show=False)

        pathcols = [child for child in ax.get_children()
                    if child.__class__.__name__ == "PathCollection"]

        modded1 = diagrams[0]
        modded1[:, 1] = diagrams[0][:, 1] - diagrams[0][:, 0]
        modded2 = diagrams[1]
        modded2[:, 1] = diagrams[1][:, 1] - diagrams[1][:, 0]
        assert len(pathcols) == 2
        np.testing.assert_array_equal(pathcols[0].get_offsets(), modded1)
        np.testing.assert_array_equal(pathcols[1].get_offsets(), modded2)

    def test_infty(self):
        diagrams = [
            np.array([[0, np.inf], [1, 1], [2, 4], [3, 5]]),
            np.array([[0.5, 3], [2, 4], [4, 5], [10, 15]])
        ]

        f, ax = plt.subplots()
        plot_diagrams(diagrams, legend=True, show=False)

        # Right now just make sure nothing breaks


class TestMatching:
    def test_bottleneck_matching(self):
        dgm1 = np.array([
            [0.1, 0.2],
            [0.2, 0.4]
        ])
        dgm2 = np.array([
            [0.1, 0.2],
            [0.3, 0.45]
        ])

        d, (matching, D) = persim.bottleneck(dgm1, dgm2, matching=True)
        persim.bottleneck_matching(dgm1, dgm2, matching, D)

    def test_plot_labels(self):
        dgm1 = np.array([
            [0.1, 0.2],
            [0.2, 0.4]
        ])
        dgm2 = np.array([
            [0.1, 0.2],
            [0.3, 0.45]
        ])

        d, (matching, D) = persim.bottleneck(dgm1, dgm2, matching=True)
        persim.bottleneck_matching(
            dgm1, dgm2, matching, D, labels=["X", "Y"])
